import createTokenMiddleware, { successCallback } from '../src/tokenMiddleware';
import sinon from 'sinon';
import expect from 'expect';
import { STORAGE_KEY } from '../src/constants';

describe('Token Middleware', () => {
  let config;
  let redirectForTokenStub;
  let renewTokenSilentAsyncStub;
  let createTokenManagerStub;
  let tokenManager;
  let getItemStub;
  let setItemStub;
  let store;
  let getStateStub;
  let next;
  let action;
  let oldStorage;
  let thenStub;

  beforeEach(() => {
    action = {type: 'SOME_ACTION'};
    config = {};
    redirectForTokenStub = sinon.stub();
    renewTokenSilentAsyncStub = sinon.stub();
    thenStub = sinon.stub();
    thenStub.returns({
      catch: sinon.stub()
    });
    renewTokenSilentAsyncStub.returns({
      catch: sinon.stub(),
      then: thenStub
    });
    tokenManager = {
      redirectForToken: redirectForTokenStub,
      expired: false,
      renewTokenSilentAsync: renewTokenSilentAsyncStub
    };
    createTokenManagerStub = sinon.stub();
    createTokenManagerStub.returns(tokenManager);
    getStateStub = sinon.stub();
    store = { getState: getStateStub };
    next = sinon.stub();

    expect(typeof(createTokenMiddleware)).toEqual('function');
    createTokenMiddleware.__Rewire__('createTokenManager', createTokenManagerStub);

    getItemStub = sinon.stub();
    setItemStub = sinon.stub();

    oldStorage = localStorage;
    localStorage = {
      getItem: getItemStub,
      setItem: setItemStub
    };
  });

  it('should return the correct middleware function', () => {
    const middleware = createTokenMiddleware(config);

    expect(typeof(middleware)).toEqual('function');
    expect(middleware.length).toEqual(1);

    let nextFunction = middleware(store);
    expect(typeof(nextFunction)).toEqual('function');
    expect(nextFunction.length).toEqual(1);

    nextFunction = nextFunction(next);
    expect(typeof(nextFunction)).toEqual('function');
    expect(nextFunction.length).toEqual(1);
  });

  it('should call the shouldValidate() function with the redux state and dispatched action', () => {
    const state = { some: 'value' };
    getStateStub.returns(state);
    next.returns(action);
    const shouldValidate = sinon.stub();
    createTokenMiddleware(config, shouldValidate)(store)(next)(action);

    expect(getStateStub.called).toEqual(true);
    expect(shouldValidate.calledWith(state, action)).toEqual(true);
  });

  it('should return the next action when token is valid', () => {
    next.returns('nextReturn');
    const middleware = createTokenMiddleware(config)(store)(next);

    const result = middleware(action);
    expect(result).toEqual('nextReturn');
    expect(next.calledWith(action)).toEqual(true);
  });

  it('should trigger the authentication flow, set the redirect url if the token is invalid & no redirect url is set', () => {
    next.returns('nextReturn');
    const middleware = createTokenMiddleware(config)(store)(next);
    tokenManager.expired = true;
    getItemStub.returns(null);

    middleware(action);

    expect(getItemStub.calledWith(STORAGE_KEY)).toEqual(true);
    expect(setItemStub.calledWith(STORAGE_KEY)).toEqual(true);
    expect(redirectForTokenStub.called).toEqual(true);
  });

  it('should not trigger the authentication flow when the redirect url is set', () => {
    getItemStub.returns('some url');
    const middleware = createTokenMiddleware(config)(store)(next);
    tokenManager.expired = true;

    middleware({});

    expect(getItemStub.called).toEqual(true);
    expect(redirectForTokenStub.called).toEqual(false);
  })

  it('should call next() with the action when it has been provided', () => {
    const middleware = createTokenMiddleware(config, null, action)(store)(next);
    tokenManager.expired = true;
    getItemStub.returns(null);

    middleware({});

    expect(next.calledWith(action)).toEqual(true);
  });

  it('should not trigger the token validation when shouldValidate returns false', () => {
    const shouldValidate = (state, action) => false;
    const middleware = createTokenMiddleware(config, shouldValidate);

    middleware({});

    expect(getItemStub.called).toEqual(false);
    expect(redirectForTokenStub.called).toEqual(false);
  });

  it('should trigger the silent refresh flow when the "silent_renew" config option is true & not trigger the default auth flow', () => {
    config = {
      silent_renew: true,
    };
    tokenManager.expired = true;
    const middleware = createTokenMiddleware(config)(store)(next);

    middleware({});

    expect(getItemStub.called).toEqual(false);
    expect(setItemStub.called).toEqual(false);
    expect(redirectForTokenStub.called).toEqual(false);
    expect(renewTokenSilentAsyncStub.called).toEqual(true);
  });

  it('should call the next middleware with the dispatched action when silent refresh is triggered', () => {
    config = {
      silent_renew: true,
    };
    tokenManager.expired = true;

    const middleware = createTokenMiddleware(config)(store)(next);

    middleware(action);
    expect(next.calledWith(action)).toEqual(true);
  });

  it('should call the then() method of silent renew when silent renew was successful', () => {
    config = {
      silent_renew: true,
    }

    tokenManager.expired = true;

    const middleware = createTokenMiddleware(config)(store)(next);

    middleware(action);

    expect(thenStub.called).toEqual(true);
  });

  it('should call the next middleware with dispatchOnSuccess object when silent renew was successful', () => {
    successCallback(next, action);

    expect(next.calledWith(action)).toEqual(true);
  });

  it('should call the next middleware with dispatchOnSuccess function when silent_renew was successful', () => {
    const stub = sinon.stub();
    stub.returns(action);

    successCallback(next, stub);

    expect(stub.called).toEqual(true);
    expect(next.calledWith(action)).toEqual(true);
  });

  it('should do nothing when dispatchOnSuccess is invalid', () => {
    successCallback(next, 'invalid');
    successCallback(next, undefined);
    successCallback(next, null);
    successCallback(next, 9001);

    expect(next.called).toEqual(false);
  });

  afterEach(() => {
    localStorage = oldStorage;
  });
})
